#include <iostream>
#include <memory>
#include <set>
#include <random>
#include <ctime>
#include <algorithm>
#include "HeapObject.h"
#include "CashFactory.h"
#include "Bank.h"

using namespace banking;

std::default_random_engine generator(static_cast<unsigned long>(time(nullptr)));
std::uniform_int_distribution<int> distribution(-100, 500);

std::string currency_to_string(CurrencyValue value) {
    if (value == CurrencyValue::USD) return "US Dollar";
    if (value == CurrencyValue::EUR) return "Euro";
    if (value == CurrencyValue::GPD) return "British Pound";
    return "Say what ?";
}

std::ostream &operator<<(std::ostream &os, Currency currency) {
    os << currency_to_string(currency.getValue());
    return os;
}

std::ostream &operator<<(std::ostream &os, Bill bill) {
    os << bill.getValue() << " " << bill.getCurrency() << " " << bill.getSerial();
    return os;
}

std::ostream &operator<<(std::ostream &os, Cash cash) {
    os << cash.getValue() << " " << cash.getCurrency();
    return os;
}

std::ostream &operator<<(std::ostream &os, const std::shared_ptr<Money> &money) {
    os << money->getValue() << " " << money->getCurrency();
    return os;
}

std::ostream &operator<<(std::ostream &os, const std::shared_ptr<Person> &person) {
    os << person->getName();
    return os;
}

std::ostream &operator<<(std::ostream &os, const std::shared_ptr<BankAccount> &bankAccount) {
    os << bankAccount->getOwner() << " " << bankAccount->getName() << ": " << bankAccount->getMoney();
    return os;
}

bool sortMoney(std::shared_ptr<Money> const &a, std::shared_ptr<Money> const &b) {
    return a->getValue() < b->getValue();
}

std::set<std::shared_ptr<Bank>> banks;
std::vector<std::shared_ptr<Money>> moneyVector;

void setup() {

    std::shared_ptr<Bank> bank1 = std::make_shared<Bank>("Bank 1");
    std::shared_ptr<Bank> bank2 = std::make_shared<Bank>("Bank 2");
    std::shared_ptr<Bank> bank3 = std::make_shared<Bank>("Bank 3");

    std::shared_ptr<Person> max = std::make_shared<Person>("Max Müller");
    std::shared_ptr<Person> marius = std::make_shared<Person>("Marius Müller");
    std::shared_ptr<Person> alex = std::make_shared<Person>("Alex Maier");


    bank1->addAccount(new BankAccount(max, "Konto 1"));
    bank2->addAccount(new BankAccount(max, "Konto 2"));
    bank1->addAccount(new BankAccount(alex, "privat"));
    bank3->addAccount(new BankAccount(alex, "geschäftlich 1"));
    bank2->addAccount(new BankAccount(alex, "geschäftlich 2"));
    bank1->addAccount(new BankAccount(alex, "geschäftlich 3"));
    bank2->addAccount(new BankAccount(marius, "Konto 1"));
    bank1->addAccount(new BankAccount(marius, "Konto 2"));
    bank3->addAccount(new BankAccount(marius, "Konto 3"));

    banks.insert(std::move(bank1));
    banks.insert(std::move(bank2));
    banks.insert(std::move(bank3));
}

void simulate() {
    auto *cashFactory = new CashFactory();

    for (auto && bank : banks) {
        std::cout << std::endl;
        std::cout << "======================== " << bank->getName() << " ========================" << std::endl;
        std::cout << std::endl;
        for (auto && bankAccount : bank->getAccounts()) {
            for (int i = 0; i < 10000; ++i) {
                int rnd = distribution(generator);
                std::shared_ptr<Bill> bill = cashFactory->printBill(rnd, CurrencyValue::USD);
                std::shared_ptr<Coin> coin = cashFactory->printCoin(rnd, CurrencyValue::USD);
                if (bill != nullptr) {
                    *bankAccount + bill;
                    moneyVector.push_back(std::static_pointer_cast<Money, Bill>(bill));
                }
                if (coin != nullptr) {
                    *bankAccount + coin;
                    moneyVector.push_back(std::static_pointer_cast<Money, Coin>(coin));
                }
            }
            std::cout << bankAccount << std::endl;
        }
    }

    for(auto && bank1 : banks) {
        for(auto && bank2 : banks) {
            if(bank1 != bank2) {
                for (int i = 0; i < 10000; ++i) {
                    int rnd = distribution(generator);
                    bank1->moveMoney(bank2, rnd);
                }
            }
        }
    }

    std::sort(moneyVector.begin(), moneyVector.end(), sortMoney);

    for(auto && money : moneyVector) {
        std::cout << money->getValue() << ", ";
    }
    std::cout << std::endl;


}

void tearDown() {
    banks.clear();
    moneyVector.clear();
}


int main(int argc, char **argv) {
    std::cout << "+++++++++++++++++++++++++ " << "setup()" << " +++++++++++++++++++++++++" << std::endl;
    setup();
    std::cout << "+++++++++++++++++++++++++ " << "simulate()" << " +++++++++++++++++++++++++" << std::endl;
    simulate();
    tearDown();


    HeapObject::assertionsHold();

    return 0;
}